import os
import requests
from dh_agent import settings
from dh_agent.tasks import get_file_extension
from dh_corpus import *
from dh_job import *
from string import Template
from PIL import Image
import base64
import json
from google.cloud import vision


image_size_limit = 9500000


def _get_filesize_kb(file_path):
    return os.path.getsize(file_path)


def _downsized_image(big_image, new_directory):
    extension = get_file_extension(big_image)
    small_image = big_image.replace(os.path.dirname(big_image), new_directory).replace(extension, "_donwsized" + extension)
    small_width = 3000
    img = Image.open(big_image)
    width_percent = (small_width / float(img.size[0]))
    small_height = int((float(img.size[1]) * float(width_percent)))
    img.thumbnail((small_width, small_height), Image.ANTIALIAS)
    img.save(small_image)
    return small_image


def _log(path, msg):
    with open(path, 'a') as log_out:
        log_out.write("{0}\n".format(msg))

# THIS CLASS AND THE FUNCTION THAT FOLLOWS ATTEMPTS TO CONVERT Google Cloud Vision JSON INTO HOCR
# THE CODE WAS COPIED FROM https://github.com/dinosauria123/gcv2hocr
class GCVAnnotation:

    templates = {
        'ocr_page': Template("""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="$lang" lang="$lang">
  <head>
    <title>$title</title>
    <meta http-equiv="Content-Type" content="text/html;charset=utf-8" />
    <meta name='ocr-system' content='Google Cloud Vision' />
    <meta name='ocr-langs' content='$lang' />
    <meta name='ocr-number-of-pages' content='1' />
    <meta name='ocr-capabilities' content='ocr_page ocr_carea ocr_line ocrx_word ocrp_lang'/>
  </head>
  <body>
    <div class='ocr_page' lang='$lang' title='bbox 0 0 $page_width $page_height'>
        <div class='ocr_carea' lang='$lang' title='bbox $x0 $y0 $x1 $y1'>$content</div>
    </div>
  </body>
</html>
    """),
        'ocr_line': Template("""
            <span class='ocr_line' id='$htmlid' title='bbox $x0 $y0 $x1 $y1; baseline $baseline'>$content
            </span>"""),
        'ocrx_word': Template("""
                <span class='ocrx_word' id='$htmlid' title='bbox $x0 $y0 $x1 $y1'>$content</span>""")
    }

    def __init__(self,
                 htmlid=None,
                 ocr_class=None,
                 lang='unknown',
                 baseline="0 -5",
                 page_height=None,
                 page_width=None,
                 content=[],
                 box=None,
                 title='',
                 savefile=False):
        self.title = title
        self.htmlid = htmlid
        self.baseline = baseline
        self.page_height = page_height
        self.page_width = page_width
        self.lang = lang
        self.ocr_class = ocr_class
        self.content = content
        self.x0 = box[0]['x'] if 'x' in box[0] and box[0]['x'] > 0 else 0
        self.y0 = box[0]['y'] if 'y' in box[0] and box[0]['y'] > 0 else 0
        self.x1 = box[2]['x'] if 'x' in box[2] and box[2]['x'] > 0 else 0
        self.y1 = box[2]['y'] if 'y' in box[2] and box[2]['y'] > 0 else 0

    def maximize_bbox(self):
        self.x0 = min([w.x0 for w in self.content])
        self.y0 = min([w.y0 for w in self.content])
        self.x1 = max([w.x1 for w in self.content])
        self.y1 = max([w.y1 for w in self.content])

    def __repr__(self):
        return "<%s [%s %s %s %s]>%s</%s>" % (self.ocr_class, self.x0, self.y0,
                                              self.x1, self.y1, self.content,
                                              self.ocr_class)

    def render(self):
        if type(self.content) == type([]):
            content = "".join(map(lambda x: x.render(), self.content))
        else:
            content = self.content
        return self.__class__.templates[self.ocr_class].substitute(self.__dict__, content=content)


def fromResponse(resp, baseline_tolerance=2):
    last_baseline = -100
    page = None
    curline = None
    if isinstance(resp, bool) and not resp:
        box = [{"x": 0, "y": 0}, {"x": 0, "y": 0}, {"x": 0, "y": 0}, {"x": 0, "y": 0}]
        page = GCVAnnotation(
            ocr_class='ocr_page',
            htmlid='page_0',
            box=box
        )
    else:
        for anno_idx, anno_json in enumerate(resp['textAnnotations']):
            box = anno_json['boundingPoly']['vertices']
            if anno_idx == 0:
                page = GCVAnnotation(
                    ocr_class='ocr_page',
                    htmlid='page_0',
                    box=box
                    )
                continue
            word = GCVAnnotation(ocr_class='ocrx_word', content=anno_json['description'], box=box)
            curline = GCVAnnotation(
                    ocr_class='ocr_line',
                    htmlid="line_%d" % (len(page.content)),
                    content=[],
                    box=box)
            page.content.append(curline)
            last_baseline = word.y1
            word.htmlid="word_%d_%d" % (len(page.content) - 1, len(curline.content))
            curline.content.append(word)
        for line in page.content:
            line.maximize_bbox()
        page.maximize_bbox()
    if not page.page_width:
        page.page_width = page.x1
    if not page.page_height:
        page.page_height = page.y1
    return page


def run(context):

    document = Document(context['connection'], id=context['job'].target_id, corpus_id=context['job'].corpus_id)
    pages = document.all_pages
    debug_log = context['log_dir'] + '/debug.txt'

    client = vision.Client()

    for page in pages:

        if os.path.exists(page.path):
            image_path = page.path

            # Google Cloud Vision API Max JSON payload is 10MB https://cloud.google.com/vision/docs/supported-files
            file_size_kb = _get_filesize_kb(image_path)
            if file_size_kb >= image_size_limit:
                _log(context['log_dir'] + '/too_big.txt', "{0}KB: {1}\n".format(file_size_kb, page.path))

                # Try to downsize image and check filesize again
                image_path = _downsized_image(image_path, context['job_dir'])
                file_size_kb = _get_filesize_kb(image_path)

                # If no luck, skip the image
                if file_size_kb >= image_size_limit:
                    continue

            _log(debug_log, "Attempting to OCR {0}: {1}KB".format(page.path, file_size_kb))

            ext = get_file_extension(image_path)
            if ext.lower() == '.tif' or ext.lower() == '.tiff':
                new_image_path = context['job_dir'] + '/' + os.path.basename(image_path).replace(ext, '.jpg')
                im = Image.open(image_path)
                im.save(new_image_path)
                image_path = new_image_path
                ext = '.jpg'

            # GET PLAIN TEXT USING GC PYTHON API
            text_path = context['job_dir'] + '/' + os.path.basename(image_path).replace(ext, '.txt')
            if not os.path.exists(text_path):
                with open(image_path, 'rb') as image_file:
                    image = client.image(content=image_file.read())
                text = image.detect_full_text().text
                with open(text_path, 'w') as text_file:
                    text_file.write(text)
            else:
                _log(debug_log, "Plain text {0} already exists.".format(text_path))

            # GET JSON USING GCV REST API
            json_path = context['job_dir'] + '/' + os.path.basename(image_path).replace(ext, '.json')
            if not os.path.exists(json_path):
                with open(image_path, 'rb') as image_file:
                    image_content = base64.b64encode(image_file.read()).decode('UTF-8')
                request_json = {
                    "requests": [
                        {
                            "image": {
                                "content": image_content
                            },
                            "features": [
                                {
                                    "type": "TEXT_DETECTION",
                                    "maxResults": 10000
                                }
                            ]
                        }
                    ]
                }
                response = requests.post(
                    url='https://vision.googleapis.com/v1/images:annotate?key=' + settings.GOOGLE_CLOUD_VISION_API_KEY,
                    data=json.dumps(request_json),
                    headers={
                        'Content-Type': 'application/json'
                    }
                )
                with open(json_path, 'w') as json_out:
                    json_out.write(response.text)
            else:
                _log(debug_log, "JSON {0} already exists.".format(json_path))

            # CONVERT GCV JSON TO HOCR XML
            try:
                xml_path = context['job_dir'] + '/' + os.path.basename(image_path).replace(ext, '.xml')
                if not os.path.exists(xml_path):
                    page = None
                    with open(json_path, 'r') as json_in:
                        gcv_json = json.load(json_in)
                        if 'responses' in gcv_json and len(gcv_json['responses']) > 0:
                            page = fromResponse(gcv_json['responses'][0])

                        with open(xml_path, 'w') as xml_out:
                            xml_out.write(page.render())
                else:
                    _log(debug_log, "XML {0} already exists.".format(xml_path))
            except:
                print(traceback.format_exc())
                print('Unable to convert GCV JSON to HOCR XML for ' + os.path.basename(image_path))


    context['job'].set_status("Job finished! Ready to transfer results")
